/*  
 Date: 10/30/2011
 Author: Honglei Liu <liuhonglei@gmail.com>
	 Jiaying Wang
	 Zhida Chen

 Description: 
 */

#include <iostream>
#include <fstream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <cassert>
#include "BWT.h"

using namespace std;

const int ALPHABET_SIZE=256;
const int BUCKET_SIZE=200;

BWT::BWT(string name):filename(name),sa_filename(name+".sa")
{
    cerr<<"Loading bwt index...";
	FILE *infile;
	infile=fopen(filename.c_str(),"rb");

	if(infile==NULL)
	{
		cerr<<"Cannot open the file:"<<filename<<endl;
		exit(1);
	}
	if(fseek(infile,0,SEEK_END)!=0)
	{
		cerr<<"Cannot locate the end of the file:"<<filename<<endl;
		exit(1);
	}

	file_len=ftell(infile);

	if(fseek(infile,0,SEEK_SET)!=0)
	{
		cerr<<"Cannot locate the start of the file:"<<filename<<endl;
		exit(1);
	}

	bwt=new char[file_len];

	unsigned read_len=fread(bwt,sizeof(char),file_len,infile);

	if(read_len!=file_len)
	{
		cerr<<"Something wrong when reading the file:"<<filename<<endl;
		exit(1);
	}

	init_struc();
	load_sa_file();
	int k=build_C_map();
	build_bucket(k);
	cerr<<"OK"<<endl;
}

void BWT::CalSA(string gram,pair<unsigned,unsigned>  &sa_range)
{
	unsigned low=0,high=file_len;

	for(int i=0;i<=gram.length()-1;++i)
	{
		int k=remap[gram[i]];
		low=static_cast<unsigned>(C[k])+occ(gram[i],low);
		high=static_cast<unsigned>(C[k])+occ(gram[i],high);
	}

	sa_range.first=low;
	sa_range.second=high;
}

void BWT::CalSA(char add_c,pair<unsigned int,unsigned int> &prev_sa,pair<unsigned int,unsigned int> &new_sa)
{
	int k=remap[add_c];
	unsigned low=prev_sa.first;
	unsigned high=prev_sa.second;

	low=static_cast<unsigned>(C[k]+occ(add_c,low));
	high=static_cast<unsigned>(C[k]+occ(add_c,high));

	new_sa.first=low;
	new_sa.second=high;
}

vector<unsigned>* BWT::GetPosList(pair<unsigned int,unsigned int> &sa_range)
{
	unsigned low=sa_range.first,high=sa_range.second;
	vector<unsigned> *pos_vec=new vector<unsigned>;

	for(size_t i=low;i<high;++i)
	{
		pos_vec->push_back(file_len-1-sa[i]);   //this position is counted from the real T's front
	}

	return pos_vec;
}

vector<unsigned>* BWT::GetPosList(string s)
{
    pair<unsigned, unsigned> sa_range;
    CalSA(s,sa_range);

	return GetPosList(sa_range);
}

void BWT::init_struc()
{
	C.resize(ALPHABET_SIZE);

	typedef vector<int>::size_type vec_size_t;

	vec_size_t vec_size=C.size();

	for(vec_size_t i=0;i!=vec_size;++i)
		C[i]=0;

	remap.resize(ALPHABET_SIZE);
	vec_size=remap.size();

	for(vec_size_t i=0;i!=vec_size;++i)
		remap[i]=-1;

	int bucket_num=(file_len+BUCKET_SIZE-1)/BUCKET_SIZE;
	bucket.resize(bucket_num);
}

void BWT::load_sa_file()
{
	int sa_value;
	ifstream infile(sa_filename.c_str());

	if(!infile)
	{
		cerr<<"Cannot open the safile:"<<sa_filename<<endl;
		exit(1);
	}

	while(infile>>sa_value)
		sa.push_back(sa_value);

	infile.close();
}

int BWT::build_C_map()
{
	size_t N=file_len-1;

	for(size_t i=0;i<=N;++i)
		++C[bwt[i]];

	int k=0;
	for(int i=0;i<ALPHABET_SIZE;++i)
	{
		if(C[i]>0)
		{
			map.push_back(i);
			remap[i]=k;
			C[k]=C[i];
			++k;
		}
	}

	return k;
}

void BWT::build_bucket(const int k)
{
	int bucket_num=bucket.size();

	vector<int> sum_bucket(k);

	for(int b=0;b<bucket_num;++b)
	{
		bucket[b].resize(k);

		for(int i=0;i<BUCKET_SIZE;++i)
		{
			unsigned pos=b*BUCKET_SIZE+i;
			if(pos>=file_len)
				break;
			++bucket[b][remap[bwt[pos]]];
		}

		for(int i=0;i<k;++i)
		{
			sum_bucket[i]+=bucket[b][i];
			bucket[b][i]=sum_bucket[i]-bucket[b][i];
		}
	}

	int sum=0;
	for(int i=0;i<k;++i)
	{
		sum+=C[i];
		C[i]=sum-C[i];
	}
}

unsigned BWT::occ(char ch,unsigned k)
{
	unsigned b=k/BUCKET_SIZE;
	unsigned l=b*BUCKET_SIZE;

	assert(bucket[b][remap[ch]]>=0);
	unsigned count=bucket[b][remap[ch]];

	for(unsigned i=l;i<k;++i)
	{
		if(bwt[i]==ch)
			++count;
	}

	return count;
}

