#include <iostream>
#include <sstream>
#include <cassert>
#include <sys/time.h>
#include <stdlib.h>

#include "Aligner.h"
#include "BWT.h"
#include "Parameters.h"
#include "pattern.h"
#include "Text.h"
#include "Tree.h"
#include "input.h"
#include "output.h"

using namespace std;

void unordered_mapTest()
{
    cout<<"testing unordered_map...";

	unordered_map<pair<unsigned int,unsigned int>, unsigned int,pair_hashh,compare_pair> testMatrix;

	testMatrix[make_pair(0,0)]=1;
	testMatrix[make_pair(1,1)]=2;
	testMatrix[make_pair(2,2)]=3;
	testMatrix[make_pair(3,3)]=4;

	assert(testMatrix.size()==4);
	//unordered_map<pair<unsigned int,unsigned int>, unsigned int,pair_hashh,compare_pair>::iterator iter;;
	//for (iter=testMatrix.begin();iter!=testMatrix.end();++iter)
	//{
	//	cout<<"["<<iter->first.first<<","<<iter->first.second<<"]"<<"->"<<iter->second<<endl;
	//}

	testMatrix.erase(make_pair(2,2));
	assert(testMatrix.size()==3);
	//for (iter=testMatrix.begin();iter!=testMatrix.end();++iter)
	//{
	//	cout<<"["<<iter->first.first<<","<<iter->first.second<<"]"<<"->"<<iter->second<<endl;
	//}

	cout<<"OK"<<endl;

}

void patternTest()
{
	cout<<"testing pattern.h....";
	string s="abcdefgh";
	Pattern p(s);
	assert(Parameters::qLen==4);
	assert(p.GetNextGram()=="abcd");
	assert(p.GetNextGram()=="bcde");
	assert(p.GetNextGram()=="cdef");
	assert(p.GetNextGram()=="defg");
	assert(p.GetNextGram()=="efgh");
	assert(p.GetNextGram()=="");

	s="abcdefghabcdefgh";
	p.SetPattern(s);
	vector<unsigned int>* poslist=new vector<unsigned int>;
	poslist->push_back(0);
	poslist->push_back(8);
	assert((*p.GetGramList("abcd"))==(*poslist));
	assert(p.GetChar(0)=='a');
	assert(p.IsCharExist(15)==true);
	assert(p.IsCharExist(16)==false);

	cout<<"OK"<<endl;

}

void bwtTest()
{
    cout<<"testing bwt...."<<endl;
    BWT bwt("../data/300k.index");

    vector<unsigned> *pos1=bwt.GetPosList("CTCTC");

    pair<unsigned,unsigned> prev_sa, sa;

    bwt.CalSA("CTCT",prev_sa);
    bwt.CalSA('C',prev_sa,sa);
    vector<unsigned> *pos2=bwt.GetPosList(sa);

	for(size_t i=0;i!=pos1->size();++i)
		assert(pos1->at(i)==pos2->at(i));

	cout<<"OK"<<endl;
}

void bwt_time_Test()
{
unsigned totalTime = 0;
struct timeval tv1, tv2;
struct timezone tz;
    cout<<"testing bwt time...."<<endl;
    BWT bwt("../data/300k.index");

gettimeofday(&tv1, &tz);

    vector<unsigned> *pos=bwt.GetPosList("TAACCCTAACCC");

gettimeofday(&tv2, &tz);
totalTime = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);

	for(size_t i=0;i!=pos->size();++i)
		cout<<"Position:"<<(*pos)[i]<<endl;

cout<<"BWT query time:"<<totalTime<<"us"<<endl;

	cout<<"OK"<<endl;
}
void alignerTest(string index_name, string query_name)
{
unsigned totalTime = 0;
struct timeval tv1, tv2;
struct timezone tz;

    cout<<"testing Aligner...."<<endl;
    string p;
    readString(p,query_name);
    Aligner aligner(index_name,p);

    cout<<endl;
    cout<<"Index File      : "<<index_name<<endl;
    cout<<"Query File      : "<<query_name<<endl;
    cout<<endl;

gettimeofday(&tv1, &tz);

    aligner.Alignment();

gettimeofday(&tv2, &tz);

    aligner.PrintParam();

totalTime = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
cout<<"Alignment time:"<<totalTime/1000<<"ms"<<endl<<endl;

    //aligner.Output();

    cout<<"OK"<<endl<<endl<<endl;
}

void cutString(const string &source, const string &target, unsigned length)
{
    string base;
	get_base_info(base,source,length);
	writeString(base,target);
}

void genQuery(const string & file_name, unsigned int file_size)
{
    srand(0);

    string base;
    readString(base,file_name);
    vector<string> s;

    for(unsigned i=0;i<1;i++)
    {
        unsigned startPos=rand()%(base.length()-file_size*1000);
        s.push_back(base.substr(startPos,file_size*1000));
    }
    stringstream ss;
    ss<<file_size;
    writeString(s, "../data/protein/query"+ss.str()+"k.txt");
}

int main(int argc, char* argv[])
{
	//unordered_mapTest();
	//patternTest();

	//bwtTest();
	//bwt_time_Test();


    //cutString("../data/chr1.fa","../data/100M.txt",100000000);

	if(argc > 3){
	    Parameters::thresholdH=atoi(argv[3]);
	    alignerTest(string(argv[1]),string(argv[2]));
	}
	else
	{
	    if(argc>2)
          alignerTest(string(argv[1]),string(argv[2]));
        else
        {
            cout<<"Syntax:  unittest <index file> <query file> [H]"<<endl;
        }
	}

    //alignerTest("./data/50M.index","./data/query/query5K.txt");

	/*for(int i=1;i<=10;i++)
        genQuery("../data/protein/P1.3M.txt",i);*/


}
